#include "chil_global.h"

int
CHILReadInformationRecord(ChilboltonFile *chilfile,RasterInfo *info)
{
#define mw( a, b ) ( a<<8 | b )

  unsigned char *ir, new_ir_format;
  short int i;

  if (chilfile->position != START_OF_IR
      && chilfile->position != START_OF_FILE ) {
    CHILPrintf("Warning: attempt to read Information Record when not at appropriate position\n");
    return END_OF_DATA;
  }

  if (fread(info->raw,sizeof(info->raw),1,chilfile->file) < 1) {
    return (chilfile->position = END_OF_DATA);
  }
  (chilfile->bytes_read) += sizeof(info->raw);
  chilfile->position = START_OF_DATA;
  ir=info->raw;

  if (ir[0] != 255) {
    CHILPrintf("Information record starts with %d, not 255: aborting\n", ir[0]);
    return (chilfile->position = END_OF_DATA);
  }

  /* First fill in the obvious fields */
  info->recording_flag = ir[1];
  info->min_gate = mw(ir[9], ir[10]);
  info->max_gate = mw(ir[11], ir[12]);
  info->threshold = mw(ir[16], ir[17]);
  info->min_angle = mw(ir[20], ir[21]);
  info->max_angle = mw(ir[22], ir[23]);
  info->central = mw(ir[24], ir[25]);
  info->system_id = ir[29]; /* This is 0 in the old format.. set correctly below */
  info->day = ir[31];
  info->month = ir[32];
  info->year = ir[33];
  info->nscans = ir[34];
  info->expt_id = ir[35];
  info->scan_velocity = ir[36];
  info->scan_type = ir[37];
  info->raster = mw(ir[39], ir[40]);
  info->sens_flag = ir[41];
  info->hour = ir[42];
  info->oper[0] = ir[43];
  info->oper[1] = ir[44];
  info->oper[2] = ir[45];
  info->minute = ir[46];
  info->second = ir[47];
  info->init_angle = mw(ir[48], ir[49]);
  info->angle_inc = mw(ir[50], ir[51]);
  info->file = mw(ir[52], ir[53]);
  info->time_stamping = ir[58];
  info->cmpl_scans = ir[59];

  if (ir[57] == 1) 
    info->range_normalised = 0;
  else
    info->range_normalised = 1;

  /* Now the tricky bit: find out what the system is and what parameters 
     are stored */
  if (info->scan_type == FIXED_93 && ir[4] == 0)
    info->cloud_data=1;
  else
    info->cloud_data=0;
  info->nrawfields=0;
  new_ir_format = info->system_id == 0 ? 0 : 1;

  if (new_ir_format) {
    /* NEW IR FORMAT (as of spring 1999) */
    for (i=0;i<7;i++) {
      if (ir[27] & (1L<<i)) {
	info->field_type[info->nrawfields++] = i+1;
      }
    }
    if (ir[27] & (1L<<7)) {
      for (i=0;i<8;i++) {
	if (ir[28] & (1L<<i)) {
	  info->field_type[info->nrawfields++] = i+8;
	}
      }
    }
    /* FIXME: temporary fix for SINGAPORE data */
    /* Actually, I don't think it's necessary...
    if (info->system_id == SINGAPORE && info->nrawfields == 4 
	&& !chilfile->is_raw) {
      info->field_type[0] = V2;
      info->field_type[1] = WIDTH_H;
      info->field_type[2] = ZH;
      info->field_type[3] = LDR;
    }
    */
    /* FIXME: There seems to be a time stamp in singapore data
       that is not indicated by the information record */
    if (info->system_id == SINGAPORE) {
      info->time_stamping = 1;
    }
  } else {
    /* OLD IR FORMAT */
    for (i=26;i<29;i++) {
      if (ir[i])
	info->field_id[info->nrawfields++] = ir[i];
    }
    /* Non-foolproof way to guess the system: */
    if (info->field_id[0] < 30) {
      info->system_id = CAMRA;
    } else if (info->field_id[0] < 40 ) {
      info->system_id = RABELAIS;
    } else if (info->field_id[0] < 50 ) {
      /* For some reason some cloud 35 GHz data uses 40 for Z... */
      if (info->cloud_data) {
	info->system_id = RABELAIS;
      }
      else {
	info->system_id = SINGAPORE;
      }
    } else if (info->field_id[0] < 60 ) {
      info->system_id = GALILEO;
    } else {
      info->system_id =UNKNOWN_SYSTEM;
    }
    /* Get field type from field ID */
    for (i=0;i<info->nrawfields;i++) {
      switch (info->field_id[i]) {
      case CAMRA_ZH:case RABELAIS_ZH:case SINGAPORE_ZH:case GALILEO_ZH:
	info->field_type[i]=ZH;
	break;
      case CAMRA_ZDR:case RABELAIS_ZDR:
	info->field_type[i]=ZDR;
	break;
      case CAMRA_LDR:case SINGAPORE_LDR:
	info->field_type[i]=LDR;
	break;
      case CAMRA_PHIDP:case RABELAIS_PHIDP:
	info->field_type[i]=PHIDP;
	break;
      case CAMRA_V:case CAMRA_VH:case RABELAIS_V:case SINGAPORE_V:
      case GALILEO_V:
	info->field_type[i]=V1;
	break;
      case CAMRA_VV:case CAMRA_V2:case RABELAIS_VV:
	info->field_type[i]=V2;
	break;
      case CAMRA_WIDTH:case SINGAPORE_WIDTH:
	info->field_type[i]=WIDTH_H;
	/*	info->field_type[++i]=WIDTH_V; */
	break;
      case CAMRA_WIDTHH:
	info->field_type[i]=WIDTH_H;
	break;
      case CAMRA_WIDTHV:
	info->field_type[i]=WIDTH_V;
	break;
      default:
	info->field_type[i]=UNKNOWN;
      }
    }
  }

  /* Get frequency from system id */
  switch (info->system_id) {
  case CAMRA:
    info->frequency=CAMRA_FREQ;
    break;
  case RABELAIS:
    info->frequency=RABELAIS_FREQ;
    break;
  case GALILEO:
    info->frequency=GALILEO_FREQ;
    break;
  case SINGAPORE:
    info->frequency=SINGAPORE_FREQ;
    break;
  default:
    info->frequency=0.0;
  } 

  /* Get the gate spacing and range offset*/
  info->range_offset=0;
  if (info->cloud_data) {
    if (new_ir_format) {
      info->gate_spacing = mw(ir[60], ir[61]);
      info->range_offset = -(255 - ir[62]) * 256 - (256 - ir[63]);
    }
    else {
      switch (info->system_id) {
      case RABELAIS:case SINGAPORE:
	info->gate_spacing=ir[14]*MIN_SPACING;
	break;
      case GALILEO:
	if (info->file < 8282)
	  info->gate_spacing=ir[14]*MIN_SPACING;
	else
	  info->gate_spacing=ir[14]*MIN_SPACING_GALILEO_NEW;
	break;
      }
    }
  }
  else {
    switch (info->system_id) {
    case CAMRA:
      switch (info->field_type[0]) {
      case ZH:case ZDR:case LDR:
	info->gate_spacing=ir[4]*ir[13]*MIN_SPACING;
	break;
      case V1:case V2:case PHIDP:
	info->gate_spacing=ir[3]*MIN_SPACING;
	break;
      }
      break;
    case RABELAIS:
      info->gate_spacing=ir[7]*ir[8]*MIN_SPACING;
      break;
    case GALILEO:
      if (info->file < 6000)
	info->gate_spacing=ir[7]*ir[8]*MIN_SPACING;
      else
	info->gate_spacing=ir[7]*ir[8]*MIN_SPACING_GALILEO_NEW;
      break;
    default: /* The following is true of SINGAPORE */
      info->gate_spacing=ir[4]*ir[13]*MIN_SPACING;
      break;
    }
  }
  if (info->gate_spacing == 0) {
    CHILPrintf("Warning: could not identify range spacing\n");
  }



  for (i=0;i<info->nrawfields;i++) {
    switch (info->field_type[i]) {
    case ZH:
      if (info->system_id == CAMRA || info->system_id == SINGAPORE) {
	info->field_function[i]=CHILzh;
      }
      else {
	info->field_function[i]=CHILzh_highprecision;
      }
      break;
    case ZDR:
      if (info->system_id == CAMRA || info->system_id == SINGAPORE) {
	info->field_function[i]=CHILzdr;
      }
      else {
	info->field_function[i]=CHILzdr_highprecision;
      }
      break;
    case LDR:
      info->field_function[i]=CHILldr;
      break;
    case PHIDP:
      if (info->file >= 5346 && info->file <= 5362) {
	info->field_function[i]=CHILphidp_old;
      }
      else if (info->file > 5362 && info->file <= 5626) {
	info->field_function[i]=CHILphidp;
      }
      else {
	info->field_function[i]=CHILphidp_new;
      }
      break;
    case V1:case V2:case WIDTH_H:case WIDTH_V:
      if (info->file >=5346 && info->file <= 5362) {
	info->field_function[i]=CHILvel_old;
      }
      else {
	info->field_function[i]=CHILvel;
      }
      break;
    default:
      info->field_function[i]=CHILraw;
    }
  }

  /* Universal format sweep modes */
  switch (info->scan_type) {
  case RHI_93:
    info->sweep_mode = UF_RHI;
    break;
  case PPI_93:
    info->sweep_mode = UF_PPI;
    break;
  case SP_93:
    info->sweep_mode = UF_COPLANE;
    break;
  case FIXED_93:
    info->sweep_mode = UF_DWELL;
    break;
  case CAL_93:
    info->sweep_mode = UF_CAL;
    break;
  case TS_93:
    info->sweep_mode = UF_MANUAL;
    break;
  default:
    CHILPrintf( "Impossible scan type (%d): aborting\n", info->scan_type);
    return (chilfile->position = END_OF_DATA);
  }

  /* often in large multi-day cloud radar files the date in the rasters 
     is wrong - use summary files */
  if (info->cloud_data == 1 && chilfile->summary_exists) {
    for (i=0;i<chilfile->nrasters;i++) {
      if (chilfile->raster[i]==info->raster) {
#ifdef DEBUG
	CHILPrintf("Fixing date\n");
#endif
	info->day=chilfile->day[i];
	info->month=chilfile->month[i];
	info->year=chilfile->year[i];
	break;
      }
    }
  }
  return START_OF_DATA;
}

/* The following is the `internal' function for initialising rays */
/* if `new' is 1 then memory is allocated for the first time, if it is 0 
   then no allocation takes place */
int
__CHILSetRay(const RasterInfo *info,int avegates,int averays,Ray *ray, uchar new)
{
  long int newsize;
  float *buffer = NULL;
  uword i;
  float start_range;

  ray->range_normalised=info->range_normalised;
  ray->correct_range=0;

  ray->utc=0.0;
  ray->last_utc=0.0;
  ray->hour_mod=0;

  ray->sigma_v_floor=0;
  ray->v_floor=0;

  ray->day=info->day;
  ray->month=info->month;
  ray->year=info->year;

  ray->raw_utc=ray->utc;
  ray->raw_day=ray->day;
  ray->raw_month=ray->month;
  ray->raw_year=ray->year;

  ray->averays=averays;
  ray->avegates=avegates;
  /* In cloud data recorded on the DSP system, max_gate was 1 too low
     - could solve using the following:

       ray->endgate = ray->ngates = (info->max_gate - info->min_gate + 1) / avegates;

     but better to use "-extragates 1" in the higher level programs */
  ray->endgate = ray->ngates = (info->max_gate - info->min_gate) / avegates;

  ray->startgate = 0;
  ray->ntotal = ray->nrawfields = info->nrawfields;

  ray->use_v1fold = ray->use_v2fold = 0;

  ray->pci_clean = 0;

  /* Work out the number of fields we intend to derive from the raw fields */
  /* 
     At the moment the only one we consider is the standard deviation of the 
     raw-ray velocity over the (avegates x averays) pulses we are averaging;
     only implemented if (avegates>1 || averays>1) 
     && raster_info->cloud_data==1
  */

  for (i = 0; i < ray->nrawfields && i < MAX_FIELDS; i++) {
    ray->field_type[i]=info->field_type[i];
    if (info->cloud_data==1 && (averays>1 || avegates>1) &&
	(ray->field_type[i]==V1 || ray->field_type[i]==V2)) {
#ifdef DEBUG
      CHILPrintf("Setting up sigma-v calculation\n");
#endif
      ray->field_type[ray->ntotal]=V_STD;
      ray->field_info[ray->ntotal]=i;
      ++(ray->ntotal);
    }
  }

  ray->zh_index = CHILFindField(ray, ZH); /* Cache it for speed! */

  newsize = ray->ngates * (sizeof(float) * (ray->ntotal+1) 
			   + 2 * sizeof(double) * ray->ntotal);

  if (new) {
#ifdef DEBUG
    CHILPrintf("Allocating space for %ld floats in memory\n",newsize/sizeof(float));
#endif

    /* This function checks for out of memory */
    buffer = (float *) CHILCalloc(newsize/sizeof(float),sizeof(float));

    if (ray->avegates > 1 || ray->averays > 1) {
      ray->npixels = (int *) CHILCalloc(ray->ngates, sizeof(int));
    }
    else {
      ray->npixels = NULL;
    }

    ray->maxsize=newsize;
   
    ray->range = (float *) buffer;
    for (i=0;i<ray->ntotal;i++) {
      ray->data[i] = ray->range + ray->ngates * (i+1);
      ray->sum[i] = ((double *) (ray->range + (ray->ntotal+1) * ray->ngates))
	+ (ray->ngates * i);
      ray->sum2[i] = ray->sum[0] + ray->ntotal * ray->ngates
	+ (ray->ngates * i);
#ifdef DEBUG
      CHILPrintf("value at 0+%d=%f\n",ray->data[i]-ray->range,*(ray->data[i]));
      CHILPrintf("value at 0+%d",((float *) ray->sum[i]) - ray->range);
#endif
    }
    /* start_range=info->range_offset+(info->min_gate+1)*info->gate_spacing; */
    start_range=((((float) info->min_gate-1)/((float) avegates))
		 +0.5) * (info->gate_spacing * avegates);
    for (i=0;i<ray->ngates;i++)
      ray->range[i]= (start_range + ((float) i) * avegates 
		      * info->gate_spacing) *0.001;
    /* Don't need to put the averaging effect into range_offset; 
       it's already in start_range*/
    ray->range_offset=info->range_offset /*+0.5*((avegates-1)*info->gate_spacing)*/ ;
    return CHIL_OK;
  }
  else {
    /* Check that new raster has same ray size */
    if (newsize != ray->maxsize) {
      CHILPrintf("Warning: re-initialised ray has different size!\n");
      return CHIL_ERROR;
    }
  }
  return CHIL_OK;
}

int
CHILFreeRay(Ray * ray)
{
  if (ray->range) {
    CHILFree((void *) ray->range);
    ray->range = NULL;
  }

  if (ray->npixels) {
    CHILFree((void *) ray->npixels);
    ray->npixels = NULL;
  }

  return CHIL_OK;
}

int
CHILReadRawRay(ChilboltonFile * chilfile,RasterInfo * info,Ray * ray)
{
  RayHeader header,tmp_header;
  TimeStamp timestamp,tmp_timestamp;
  short int tmp_buffer[CHIL_RAYBUFFERSIZE];
  unsigned char signal[2];
  unsigned short nwords;

  /* Read and interpret header */
#ifdef DEBUG
  CHILPrintf("Reading ray header: ");
#endif
  if (fread(signal,sizeof(char),2,chilfile->file) < 2) {
    return CHIL_ERROR;
  };
  (chilfile->bytes_read) += 2;
  if (chilfile->is_raw) {
    nwords = signal[0] | signal[1] << 8;
  }
  else {
    nwords = signal[1] | signal[0] << 8;
  }
#ifdef DEBUG
  CHILPrintf("%hu words in ray\n",nwords);
#endif

  /* if (!(nwords == END_OF_SCAN || nwords == END_OF_RASTER
     || nwords == RASTER_ABORTED)) */
  /* Well, there's not much point in testing for all the `special' values
     - we just want to avoid segfaults by checking that the data won't
     overrun the buffer */
  if (nwords <= CHIL_RAYBUFFERSIZE) {

    /* NOTE: this code can read raw data (little endian) as well as processed
       (big endian), and also be compiled for either architecture,
       so we must be careful...*/
#ifdef __IS_I386_ARCH__
#define swap_bytes (! chilfile->is_raw)
#else
#define swap_bytes (chilfile->is_raw)
#endif
    if (swap_bytes) {
      if (fread(&tmp_header,sizeof(header),1,chilfile->file) < 1) {
	return CHIL_ERROR;
      };
      swab((void *) &tmp_header,(void *) &header,sizeof(header));
    }
    else {
      if (fread(&header,sizeof(header),1,chilfile->file) < 1) {
	return CHIL_ERROR;
      };
    };
    (chilfile->bytes_read) += sizeof(header);
    if (!chilfile->is_raw) {
      nwords -= 3;
    }
    ray->elevation=header.elevation / 60.0;
    ray->azimuth=header.azimuth / 60.0;

    if (info->time_stamping) {
      if (swap_bytes) {
	if (fread(&tmp_timestamp,sizeof(timestamp),1,chilfile->file) < 1) {
	  return CHIL_ERROR;
	};
	swab((void *) &tmp_timestamp,(void *) &timestamp,sizeof(timestamp));
      }
      else {
	if (fread(&timestamp,sizeof(timestamp),1,chilfile->file) < 1) {
	  return CHIL_ERROR;
	};
      }
      (chilfile->bytes_read) += sizeof(timestamp);
      if (!chilfile->is_raw) {
	nwords -= 2;
      }
      ray->last_utc=ray->utc;
      ray->utc=((double) timestamp.mins_since_midnight)/60.0 
	+ ((double) timestamp.centisecs)/360000.0;
      /* The small value is because I've had problems comparing floats
	 that are equal */
      if (ray->utc < ray->last_utc - 0.0003) {
	CHILIncrDate(&(ray->day),&(ray->month),&(ray->year));
      }
    }
    else if (info->cloud_data) {
      /* in the older cloud data the time is encoded into the redundant
	 az and el */
      ray->last_utc=ray->utc;
      ray->utc=info->hour+ray->azimuth + ray->elevation/60.0 + ray->hour_mod;
      if (ray->utc < ray->last_utc - 0.0003) {
	if (ray->utc >=23) {
	  /* next day! */
	  ray->utc -=23.0;
	  ray->hour_mod = -(info->hour);
	  CHILIncrDate(&(ray->day),&(ray->month),&(ray->year));
	}
	else {
	  ray->utc +=1;
	  ray->hour_mod++;
	}
      }
      ray->elevation=90.0;
      ray->azimuth=0.0;
    }
    else {
      /* No timestamp! */
      ray->utc = info->hour + info->minute/60.0 + info->second/3600.0;
    }
   
    /* Now read the ray data itself */
    if (swap_bytes) {
      if (fread(tmp_buffer, sizeof(short), nwords,chilfile->file) < nwords) {
	return CHIL_ERROR;
      }
      swab((void *) &tmp_buffer,(void *)  &ray->buffer, nwords * 2);
    } 
    else {
      if (fread(ray->buffer, sizeof(short), nwords, chilfile->file) < nwords) {
	return CHIL_ERROR;
      }
    }
    (chilfile->bytes_read) += nwords * 2;
    chilfile->position = START_OF_RAY;
  }
  else /* Required buffer size too big - probably an abort code */ {
    if (nwords == END_OF_RASTER || nwords == RASTER_ABORTED) {
      CHILNextRaster(chilfile,info);
    }
    if (nwords == END_OF_SCAN || nwords == STATUS_OK 
	|| nwords == STATUS_ABORTED || nwords == STATUS_NC) {
      chilfile->position=START_OF_IR;
      chilfile->iscan++;
    }
    return CHIL_ERROR;
  }
  ray->correct_range=0;
  ray->nez=0.0;
  return CHIL_OK;
}

int
CHILReadRay(ChilboltonFile * chilfile,RasterInfo * info,Ray * ray)
{
  int status;
  float (*func) (short);
  float (*zh_func) (short) = NULL;
  float utc_mod=0;
  unsigned short int field, gate, day, month, year, first_ray=1;
  int zh_index = CHILFindZH(ray);  

  if (zh_index >= 0) {
    zh_func = info->field_function[zh_index];
  }

  if (ray->avegates == 1 && ray->averays == 1) {
    /* NO AVERAGING CASE */
    /* If the ray is corrupted due to the strange PCI card bug (?) then get another */
    do {
#ifdef DEBUG
      CHILPrintf("Reading unaveraged ray\n");
#endif
      if ((status=CHILReadRawRay(chilfile,info,ray)) != CHIL_OK ) {
	return status;
      }
#ifdef DEBUG
      CHILPrintf("...got raw data\n");
#endif
    } while (ray->pci_clean && CHILIsPCIGlitched(info, ray));

    if (zh_index < 0) {
      /* No Zh field so definitely not coherent processing */
      for (field = 0; field < ray->nrawfields; field++) {
	func=info->field_function[field];
	for (gate=0;gate < ray->ngates;gate++) {
	  *(ray->data[field]+gate) = func(ray->buffer[gate * (ray->nrawfields)
						     + field]);
	}
      }
    }
    else {
      for (gate = 0; gate < ray->ngates; gate++) {
	if (zh_func(ray->buffer[gate * (ray->nrawfields) + zh_index])
	    < MIN_ZH) {
	  /* Coherent processing says there is no signal here */
	  for (field = 0; field < ray->nrawfields; field++) {
	    *(ray->data[field]+gate) = chil_no_signal_value;
	  }
	}
	else {
	  for (field = 0; field < ray->nrawfields; field++) {
	    func = info->field_function[field];
	    *(ray->data[field]+gate) = func(ray->buffer[gate * (ray->nrawfields)
						       + field]);
	  }
	}
      }
    }
  }
  else {
    int i,j;
    /* AVERAGING CASE - first set everything to zero*/

    ray->sum_elevation=0;
    ray->sum_azimuth=0;
    ray->sum_utc=0;

    /* Remember that utc is the *mean* utc of an averaged ray, so when that
       ray spans midnight the `mean' date of the ray can be `yesterday' and
       unless care is taken the next ray will think it's yesterday too... */
    ray->utc=ray->raw_utc;
    ray->day=ray->raw_day;
    ray->month=ray->raw_month;
    ray->year=ray->raw_year;

    for (field=0;field < ray->ntotal;field++) {
      for (gate=0;gate < ray->ngates;gate++) {
	/*           *(ray->data[field]+gate) = 0;*/
	*(ray->sum[field]+gate) = 0;
	*(ray->sum2[field]+gate) = 0;
      }
    }

    for (gate = 0; gate < ray->ngates; gate++) {
      ray->npixels[gate] = 0;
    }

#ifdef DEBUG
    CHILPrintf("Reading averaged ray\n");
#endif
    /* Loop through the high-resolution rays */
    /* Skip any glitched rays caused by a PCI card bug */
    for (i = 0; i < ray->averays; i++) {
      do {
	if ((status=CHILReadRawRay(chilfile,info,ray)) != CHIL_OK ) {
	  return status;
	}
      } while (ray->pci_clean && CHILIsPCIGlitched(info, ray));

      if (first_ray) {
	day=ray->day;
	month=ray->month;
	year=ray->year;
	first_ray=0;
      }
      else if (ray->last_utc>ray->utc) {
	utc_mod = 24.0;
      }
      ray->sum_elevation += ray->elevation;
      ray->sum_azimuth += ray->azimuth;
      ray->sum_utc += (ray->utc + utc_mod);

      /* Loop through the averaged output gates */
      for (gate = 0; gate < ray->ngates; gate++) {

	/* Loop through the high-resolution gates within each output gate */
	for (j = gate*ray->avegates; j<(gate+1)*ray->avegates; j++) {
	  /* Check if coherent system has already rejected data */
	  if (zh_index >= 0) {
	    float val;
	    if ((val = zh_func(ray->buffer[j * (ray->nrawfields) + zh_index]))
		< MIN_ZH) {
	      /* Coherent system: no signal found! Skip pixel */
	      continue;
	    }
	  }
	  ++(ray->npixels[gate]);

	  /* loop through the raw fields */
	  for (field = 0; field < ray->nrawfields; field++) {
	    func = info->field_function[field];

	    /* if Z or LDR, convert from log to linear before averging */
	    if (ray->field_type[field] == ZH || ray->field_type[field] == LDR) {
	      *(ray->sum[field]+gate) 
		+= pow(10,(double) 0.1
		       * func(ray->buffer[j * (ray->nrawfields) + field]));
	    }
	    /* if V1 or V2 and we know the folding velocity, do I & Q sumations */
	    else if (ray->field_type[field] == V1 && ray->use_v1fold) {
	      double angle = func(ray->buffer[j*(ray->nrawfields)+field])
		* M_PI / ray->v1fold;
	      *(ray->sum[field]+gate)  += (double) cos(angle);
	      *(ray->sum2[field]+gate) += (double) sin(angle);
	    }
	    else if (ray->field_type[field] == V2 && ray->use_v2fold) {
	      double angle = func(ray->buffer[j*(ray->nrawfields)+field])
		* M_PI / ray->v2fold;
	      *(ray->sum[field]+gate)  += (double) cos(angle);
	      *(ray->sum2[field]+gate) += (double) sin(angle);
	    }
	    /* otherwise convert in native form */
	    else {
	      *(ray->sum[field]+gate)
		+= (double) func(ray->buffer[j * (ray->nrawfields) + field]);
	    }
	  } /* Done raw fields */

	  /* next we do the derived fields (if there are any) */
	  for (field = ray->nrawfields; field < ray->ntotal; field++) {
	    /* standard deviation of velocity is the only one we know what to
	       do with... */
	    if (ray->field_type[field] == V_STD) {
	      double tmp_double;
	      func=info->field_function[ray->field_info[field]];
	      tmp_double = (double)
		func(ray->buffer[j * (ray->nrawfields) + ray->field_info[field]]);
	      *(ray->sum[field]+gate) += (tmp_double * tmp_double);
	    }
	  }
	} /* high-resolution gates */
      } /* averaged gates */
    } /* high-resolution rays */

    ray->raw_utc = ray->utc;
    ray->raw_day = ray->day;
    ray->raw_month = ray->month;
    ray->raw_year = ray->year;

    /* Work out averaged values from summation */
    ray->elevation = ray->sum_elevation / ray->averays;
    ray->azimuth = ray->sum_azimuth / ray->averays;
    ray->utc = ray->sum_utc / ray->averays;
    if (ray->utc > 24) {
      CHILIncrDate(&day,&month,&year);
      ray->utc -= 24;
    }

    ray->day = day;
    ray->month = month;
    ray->year = year;

    /* Loop through all the gates, calculating mean values from the sumations */
    for (gate = 0; gate < ray->ngates; gate++) {
      if (ray->npixels[gate] == 0) {
	/* Coherent system: no signal in any of the averaged pixels! */
	for (field = 0; field < ray->ntotal; field++) {
	  *(ray->data[field]+gate) = chil_no_signal_value;
	}
      }
      else {
	for (field = 0; field < ray->ntotal; field++) {
	  if (ray->field_type[field] == ZH) {
	    /* Empty pixels count as Z = 0 mm6m-3,
	       so divide by the total number of pixels */
	    /* Using logarithmic averaging */
	    *(ray->data[field]+gate) = 10.0 *
	      log10(*(ray->sum[field]+gate) / (ray->avegates * ray->averays));
	  }
	  else if (ray->field_type[field] == LDR) {
	    /* Empty pixels don't count at all,
	       so divide by the number of cloudy pixels */
	    /* Using logarithmic averaging */
	    *(ray->data[field]+gate) = 10.0 *
	      log10(*(ray->sum[field]+gate) / ray->npixels[gate]);
	  }
	  else if (ray->field_type[field] == V1 && ray->use_v1fold) {
	    *(ray->data[field]+gate) = ray->v1fold * M_1_PI
	      * atan2(*(ray->sum2[field]+gate), *(ray->sum[field]+gate));
	  }
	  else if (ray->field_type[field] == V2 && ray->use_v2fold) {
	    *(ray->data[field]+gate) = ray->v2fold * M_1_PI
	      * atan2(*(ray->sum2[field]+gate), *(ray->sum[field]+gate));
	  }
	  else if (ray->field_type[field] == V_STD) /* A `derived' field */ {
	    /* Before standard deviations were calculated from 2 or
	       more cloudly samples, now only calculate it if every
	       sample in the average is cloudy */
	    if (ray->npixels[gate] < (ray->avegates * ray->averays)) {
	      *(ray->data[field]+gate) = chil_no_signal_value;
	    }
	    else {
#ifdef DEBUG
	      CHILPrintf("Finalising sigma-v\n");
#endif
	      *(ray->data[field]+gate) = (float) 
		sqrt((*(ray->sum[field] + gate) / ray->npixels[gate] )
		     - (double) (*(ray->data[ray->field_info[field]] + gate)
				 * *(ray->data[ray->field_info[field]] + gate)));
	    }
	  }
	  else { 
	    /* Ordinary field - note that we divide by the number
	       of cloudy pixels */
	    *(ray->data[field]+gate) = *(ray->sum[field]+gate)
	      / ray->npixels[gate];
	  }
	} /* Loop over the different fields */
      }
    } /* Loop over the gates */
  }
  return CHIL_OK;
}

int
CHILSetFoldingVelocities(Ray *ray, float v1fold, float v2fold)
{
  if (v1fold > 0.0) {
    ray->v1fold = v1fold;
    ray->use_v1fold = 1;
  }
  if (v2fold > 0.0) {
    ray->v2fold = v2fold;
    ray->use_v2fold = 1;
  }
  return CHIL_OK;
}
